###################################################################
## Implementation of TSIV estimators and Generalized Hausman test##
###################################################################

# Author: Wei Li
# 
# ref: Optimal Linear Instrumental Variable Approximations
#      Escanciano, J.C. and Li, W., Journal of Econometrics, 1, 223-246, 2021.

library(MASS)
library("splines")
#library(gurobi)
options(scipen=999)

trace <- function(A) {
  n <- dim(A)[1] 
  tr <- 0 
  
  for (k in 1:n) {
    l <- A[k,k]
    tr <- tr + l
  }
  return(tr[[1]])
}

g_hat_f<-function(y,x,z,k1,q1,k2,q2,lambda){ 
  n<-length(y)
  K=q1+k1;
  knots1<-seq(min(x),max(x),length=k1+2)
  bdknots1<-knots1[c(1,length(knots1))]
  inknots1<-knots1[2:(length(knots1)-1)]
  
  J=q2+k2;
  knots2<-seq(min(z),max(z),length=k2+2)
  bdknots2<-knots2[c(1,length(knots2))]
  inknots2<-knots2[2:(length(knots2)-1)]
  
  Bx<-bs(x, knots=inknots1, Boundary.knots=bdknots1, degree=q1-1, intercept=TRUE) 
  Bz<-bs(z, knots=inknots2, Boundary.knots=bdknots2, degree=q2-1, intercept=TRUE)	
  
  p_hat<-Bz%*%ginv(t(Bz)%*%Bz)%*%t(Bz)%*%Bx  # nrow=n, ncol=ncol(Bx)
  P2n<-(t(Bx)%*%Bx)/n
  B_lambda<-(t(p_hat)%*%(p_hat)/n+lambda*P2n)
  B_lambda_inv<-ginv( B_lambda )
  
  Gn<- (t(p_hat)%*%y)/n  # nrow=ncol(Bx)
  g_hat<- (Bx%*%t(B_lambda_inv)%*%Gn) # n by 1
  return(g_hat)
}


GCV_TSIV<-function(y,x,z,k1,q1,k2,q2){ 

  n<-length(y)
  K=q1+k1;
  knots1<-seq(min(x),max(x),length=k1+2)
  bdknots1<-knots1[c(1,length(knots1))]
  inknots1<-knots1[2:(length(knots1)-1)]
  
  J=q2+k2;
  knots2<-seq(min(z),max(z),length=k2+2)
  bdknots2<-knots2[c(1,length(knots2))]
  inknots2<-knots2[2:(length(knots2)-1)]
  
  Bx<-bs(x, knots=inknots1, Boundary.knots=bdknots1, degree=q1-1, intercept=TRUE) 
  Bz<-bs(z, knots=inknots2, Boundary.knots=bdknots2, degree=q2-1, intercept=TRUE)	
  
  q_hat<-Bx%*%ginv(t(Bx)%*%Bx)%*%t(Bx)%*%Bz
  Q2n<-(t(Bz)%*%Bz)/n
  
  grid<-seq(.001, .1, length.out=100)
  gcv_vec<-numeric(0)
  for (k in 1:100){
    lambda<-grid[k]
    A_lambda<-(t(q_hat)%*%(q_hat)/n+lambda*Q2n)
    A_lambda_inv<-ginv( A_lambda )
    Dn<-(t(q_hat)%*%x)/n
    
    h_hatX<-t(Dn)%*%A_lambda_inv%*%((t(Bz)%*%x)/n)
    h_hatX_inv<-ginv(h_hatX)
    beta_tsiv<-as.vector(h_hatX_inv%*%(t(Dn))%*%A_lambda_inv%*%((t(Bz)%*%y)/n))
    h_hat<-(Bz%*%A_lambda_inv%*%Dn) # n by 1
    
    res<-y-x*beta_tsiv
    L_lambda<- x%*% (h_hatX_inv /n)%*%t(Dn)%*%A_lambda_inv%*%t(Bz)
    gcv<-(sum(res^2)/n)/(1- (trace(L_lambda)/n))^2
    gcv_vec<-c(gcv_vec, gcv)
  }
  
  lambda_gcv<-grid[which.min(gcv_vec)]
  
  A_lambda<-(t(q_hat)%*%(q_hat)/n+lambda_gcv*Q2n)
  A_lambda_inv<-ginv( A_lambda )
  Dn<-(t(q_hat)%*%x)/n
  
  h_hatX<-t(Dn)%*%A_lambda_inv%*%((t(Bz)%*%x)/n)
  h_hatX_inv<-ginv(h_hatX)
  beta_tsiv<-as.vector(h_hatX_inv%*%(t(Dn))%*%A_lambda_inv%*%((t(Bz)%*%y)/n))
  h_hat<-(Bz%*%A_lambda_inv%*%Dn) # n by 1
  g_hat<-g_hat_f(y,x,z,k2,q2,k1,q1,lambda_gcv)
  
  mn<- (matrix(y)-matrix(x)%*%beta_tsiv)*h_hat-(g_hat-matrix(x)%*%beta_tsiv)*(h_hat-matrix(x))
  V_hat_tsiv<-ginv((t(h_hat)%*%x)/n)*mean(mn^2)*ginv((t(h_hat)%*%x)/n)
  stdv_tsiv<-sqrt(V_hat_tsiv/n) 
  ci95_tsiv<- c(beta_tsiv-qnorm(.975)*stdv_tsiv, beta_tsiv+qnorm(.975)*stdv_tsiv)  
  
  Hhat<-cbind(rep(1,n), h_hat)
  uhat<-x-Hhat%*%(ginv(t(Hhat)%*%Hhat))%*%t(Hhat)%*%x
  reg<-cbind(rep(1,n),x)
  reg<-cbind(reg,uhat)
  A<-n*ginv(t(reg)%*%reg)
  beta_res<-A%*%(t(reg)%*%y)/n
  resid<-y-reg%*%beta_res
  L<-diag(as.vector(resid)^2)
  omega<-(A%*%(t(reg)%*%L%*%reg)%*%A)/n
  Htest<-n*(beta_res[3,1]^2)/omega[3,3]	
  
  return(list(beta_tsiv=beta_tsiv,ci95_tsiv=ci95_tsiv, Htest=Htest, h_hat=h_hat, g_hat=g_hat, lambda_gcv=lambda_gcv))
}


dgp7<-function(nobs,p,gamm,rho){
  Sigma <- matrix(c(1,gamm,gamm,1),2,2)
  datxd<-mvrnorm(n = nobs, rep(0, 2), Sigma)
  x<-datxd[,1]
  d<-datxd[,2]
  H1x<-x
  H2x<-x^2-1
  H3x<-x^3-3*x
  v<-x-gamm*d
  rhoeps<- rho/(1-gamm^2)
  err<-rnorm(n=nobs,0,1)
  eps<-rhoeps*v+err
  
  if (p==1) {
    z<-d  
    g<-H1x
  } else if (p==2) {
    z<-d^3
    g<-H1x+H2x
  } else if (p==3) {
    z<-exp(d)/(1+exp(d))
    g<-H1x+H2x+H3x
  }
  
  y<-g+eps
  truth<-1
  return(list(dat=data.frame(x=x,y=y,z=z),truth=truth))
}

dat<-dgp7(nobs=2000,p=1,gamm=.2, rho=0)

dat<-dgp7(nobs,p, gamm, rho)
x<-dat$dat$x
y<-dat$dat$y
z<-dat$dat$z

k1=4; q1=3; k2=4; q2=3

TSIV_res<-GCV_TSIV(y,x,z,k1,q1,k2,q2)
TSIV_res$beta_tsiv
TSIV_res$ci95_tsiv
TSIV_res$Htest
